"""
Use MapReduce (Hadoop) style programming to count the most common letter
unigrams (single characters) and bigrams (character pairs) in a text.
See: http://en.wikipedia.org/wiki/Bigram

Hadoop assumes that the input data is a text file. This is convenient,
because text files can be easily broken into chunks at individual lines
and distributed to many computers. It then roughly operates in this fashion:

MAP STEP: runs a developer-provided function on one line at a time of the input
files. The Map step is therefore generally used to convert the data represented
in the line into one or more key/value pairs. Normally, we could think of
"key/value" pairs as a dictionary, but Hadoop lets us have duplicate keys.
So, in essence, if we're writing code for Hadoop (which is normally done in the
Java programming language, but we're using Python for familiarity), we want
to write a map function that will accept a line of text and output a dictionary.

SHUFFLE STEP: As key/value pairs come in from the map step, hadoop sorts them
and redirects all values that have the same key to the same computer to run
the Reduce step. The map step must finish completely before moving on to
the Reduce step.

REDUCE STEP: Another developer-provided function, the reducer is given a
key and a list of all of the values for that key by Hadoop. It then
performs some computation and outputs a *single* value for that key.

SORT STEP: Hadoop will generally sort the final output data.

It is not unusual to have to run several Map/Reduce steps on a big data set
before receiving the output you want. In our example below, we'll only need a
single Map and single Reduce step to complete our goal.

See more information on MapReduce applications, of which Hadoop is a part, here:
http://en.wikipedia.org/wiki/MapReduce

For a good exercise, consider how you could modify the code below to count
unigram WORDS and bigram WORDS. In other words, instead of counting individual
characters or character pairs, counting whole words and counting word pairs.

Dan Armendariz, Dan Bradley
danallan@cs.harvard.edu, dbradley@cs50.harvard.edu
"""

# itemgetter will assist in sorting
from operator import itemgetter

def unigram_mapper(line):
    """Given a line of text, remove extraneous characters and output a key/value
    pair for each character. The character will be the key and the value will
    be 1.

    For example, given the string "Hello, world" we will output:
    [("h",1), ("e",1), ("l",1), ("l",1), ("o",1), ("w",1), ("o",1), ("r",1),
     ("l",1), ("d",1)]
    """
    output = []

    # perform the step for every character in the string
    for char in line:
        # skip the character if it is not a letter
        if not char.isalpha():
            continue

        # create our keyvalue pair, converting the letter to lower case
        keyvalue_pair = (char.lower(), 1)
        output.append(keyvalue_pair)

    return output


def bigram_mapper(line):
    """Given a line of text, remove extraneous characters and output a key/value
    pair for each bigram. The bigram will be the key and the value will be 1.

    For example, given the string "Hello, world" we will output:
    [("he",1), ("el",1), ("ll",1), ("lo",1), ("wo",1), ("or",1), ("rl",1),
     ("ld", 1)]
    """
    output = []

    # iterate over each character in the line
    for i in range(len(line)-1):
        # fetch a bigram by splicing the line at i to grab 2 neighboring chars
        bigram = line[i:i+2]

        # skip the bigram if either character isn't a letter
        if not bigram.isalpha():
            continue

        # create the key/value pair, and ensure the bigram is lower case
        keyvalue_pair = (bigram.lower(), 1)
        output.append(keyvalue_pair)

    return output


def count_reducer(key, values):
    """Given a list of values for a key, we'll return the sum of all values."""
    total = 0
    for value in values:
        total += value
    return (key, total)


def hadoop(file, mapper, reducer):
    """A higher-order function that accepts a mapper and a reducer and
    performs Hadoop/MapReduce-style processing on a given file.

    Essentially, this code (in a very simple way) performs on a single computer
    what Hadoop and other MapReduce implementations do across a wide number
    of computers.
    """
    # a structure to store the final output
    data = []

    # MAP STEP
    pairs = []
    with open(file) as f:
        # iterate over every line in the opened file
        for line in f:
            # ask our map step to process the line
            output = mapper(line)

            # store the newly computed key/value pairs in our `pairs` structure
            pairs.extend(output)

    # SHUFFLE STEP: aggregate all values for a given key
    collapsed = {}
    for pair in pairs:
        # split the pair into its key and value
        key, value = pair

        # if the key doesn't exist yet in our dictionary, add it
        if collapsed.get(key, None) is None:
            collapsed[key] = []

        # append the value to the list of values for a given key
        collapsed[key].append(value)


    # REDUCE STEP: provide all of the values for a key to the reducer
    for key, values in collapsed.iteritems():
        # the key is different; run the reduce step on our accumulated vals
        output = reducer(key, values)
        data.append(output)

    # SORT STEP: sort the output by values in descending order
    data.sort(key=itemgetter(1), reverse=True)

    # all done!
    return data

# Using "Hadoop" and our reducer and unigram mapper, count the number of
# instances each character appears in Carroll's "Alice and Wonderland"
print "Unigram character counts:"
result = hadoop("aliceandwonderland.txt", unigram_mapper, count_reducer)
for count in result:
    print count

print

# On to bigrams. Our reducer is so simple we can use the same one for both!
print "Bigram character counts:"
result = hadoop("aliceandwonderland.txt", bigram_mapper, count_reducer)

# there are many bigrams, so we'll only print the first 20
for count in result[0:10]:
    print count